from typing import List, Tuple, Dict, Union, Optional
from xml.etree.ElementTree import Element


class ColumnNotFound(Exception):
    def __init__(self, msg):
        super(ColumnNotFound, self).__init__()
        self.msg = msg

    def __repr__(self):
        return f"ColumnNotFound {self.msg}"


class Reader(object):
    @property
    def init_params(self):
        return {"reader_type": self.reader_type, "keys": self._keys, "column_marker": self.column_marker}

    def __init__(self, reader_type: str, keys: List[Tuple[str, Optional[str]]], column_marker: str = None):
        self.map_to: Dict[Union[int, str], str] = {}
        self._keys = keys
        self.column_marker: str = column_marker
        self.reader_type: str = reader_type
        self._header: List[str] = []

        self._reverse_map = {}

        if self.reader_type == "order":
            self.map_to = {int(key): mapped for key, mapped in keys if mapped}
            self._header = [
                self.map_to.get(item, "UNK"+str(item))
                for item in range(max(self.map_to.keys())+1)
            ]
        elif self.reader_type == "explicit":
            self.map_to = {key: mapped or key for key, mapped in keys}

    def get_column(self, line: List[str], column: str, raise_on_none: bool = False) -> str:
        """ Given a list of tokens, chez if we have it in our header"""
        if column in self._reverse_map:
            if self._reverse_map[column] in line:
                return line[self._reverse_map[column]]

        for pos, (value, key) in enumerate(zip(line, self.header)):
            if key == column:
                self._reverse_map[column] = pos
                return value

        if raise_on_none:
            raise ColumnNotFound(f"Column not found for Line=`{'    '.join(line)}`")

    @property
    def has_header(self):
        return self.reader_type == "explicit"

    @classmethod
    def from_xml(cls, xml_node: Element, column_marker: Optional[str] = None, default: Optional["Reader"] = None) -> "Reader":
        reader_type = xml_node.attrib["type"]
        if reader_type == "default":
            if not default:
                raise Exception("Not default reader was passed but one was asked by the configuration.")
            params = default.init_params
            params["column_marker"] = column_marker
            return cls(**params)
        _map = [(key.text.strip(), key.get("map-to")) for key in xml_node.findall("./key") if key.text]
        return cls(reader_type=reader_type, keys=_map, column_marker=column_marker)

    def set_header(self, line: str) -> List[str]:
        self._header = [
            self.map_to[key]
            for key in line.strip().split(self.column_marker)
            if key and key in self.map_to
        ]
        return self._header

    @property
    def header(self):
        return self._header

    def __repr__(self):
        return "<Reader type='{}' keys=[{}] from=[{}] />".format(
            self.reader_type,
            ", ".join(self.map_to.values()),
            ", ".join(list(map(str, self.map_to.keys())))
        )



from argparse import ArgumentParser
import os
import shutil
from typing import Optional

from .configs import ProtogenieConfiguration
from .dispatch import split_files, files_from_memory, glue
from .cli_utils import check_ratio

import click


@click.group()
def main():
    """Protogeneia is a tool to preprocess and harmonize datasets in NLP tasks. Might be useful for other stuff too
    as long as you have TSV/CSV ;)"""
    pass


@main.command("get-scheme")
@click.argument("dest", type=click.Path(file_okay=True, dir_okay=True), default="./scheme.rng")
def cli_scheme(dest):
    """Copy the schema file to [WHERE]"""
    import os
    import shutil
    here = os.path.abspath(os.path.dirname(__file__))
    schema = os.path.join(here, "schema.rng")
    shutil.copy(schema, dest)
    click.echo("Copied to {}".format(os.path.abspath(dest)))


@main.command("build")
@click.argument("file", type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option("--output", default="./output", type=str, help="Directory where the output should be built")
@click.option("-n", "--no-split", default=False, is_flag=True, help="Does not apply splitting to files")
@click.option("-c", "--clear", default=False, is_flag=True, help="Clear the output directory")
@click.option("-t", "--train", "train", default=0.8, type=float, help="Ratio (on 1) of data to use for training")
@click.option("-d", "--dev", "dev", default=0., type=float, help="Ratio (on 1) of data to use for dev set")
@click.option("-e", "--test", "test", default=0.2, type=float, help="Ratio (on 1) of data to use for test set")
@click.option("-v", "--verbose", default=False, is_flag=True, help="Print text level stats")
def cli_build(file, output="./output", no_split=False, clear=False, train=0.8, dev=.0, test=0.2, verbose=False):
    """ Uses [FILE] to split and pre-process a training corpus for NLP Tasks. File should follow the schema, see
    protogeneia get-scheme"""

    if clear:
        confirm_message = "Are you sure you want to remove data in {} ? [y/n]\t>\t".format(output)

        if click.confirm(confirm_message):
            click.echo("\tRemoving data in {}".format(output))
            shutil.rmtree(output, ignore_errors=True)
        else:
            print("\tData were not removed")
    if no_split:
        train, test, dev = 1, 0, 0

    dispatch(
        config=file,
        train=train,
        test=test,
        dev=dev,
        output_dir=output,
        verbose=verbose
    )


@main.command("rebuild")
@click.argument("file", type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.argument("memory", type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option("--output", default="./output", type=str, help="Directory where the output should be built")
@click.option("-c", "--clear", default=False, is_flag=True, help="Clear the output directory")
@click.option("-d", "--dev", "dev", default=0., type=float,
              help="[New file only] Percentage of data to use for dev set")
@click.option("-e", "--test", "test", default=None, type=float,
              help="[New file only] Percentage of data to use for test set")
def cli_rebuild(file, memory, output, clear=False, dev=.0, test=0.2):
    """Given [MEMORY] file, uses [FILE] config file to generate a new corpus

    This method detects new files and treat them if --test and --dev are given
    """

    if clear:
        confirm_message = "Are you sure you want to remove data in {} ? [y/n]\t>\t".format(output)

        if click.confirm(confirm_message):
            click.echo("\tRemoving data in {}".format(output))
            shutil.rmtree(output, ignore_errors=True)
        else:
            print("\tData were not removed")

    from_memory(
        config=file,
        memory_file=memory,
        test_ratio=test,
        dev_ratio=dev,
        output_dir=output
    )


@main.command("concat")
@click.argument("config", type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.argument("output",  type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.option("-r", "--reduce", type=click.FLOAT, default=None, help="Reduce training corpus to the given percentage")
@click.option("-p", "--prefix", type=click.STRING, default="", help="Prefix for output files")
@click.option("-v", "--verbose", default=False, is_flag=True, help="Print text level stats")
def cli_concat(config, output, reduce, verbose, prefix):
    """Given [MEMORY] file, uses [FILE] config file to generate a new corpus

    This method detects new files and treat them if --test and --dev are given
    """
    concat(config, output, verbose=verbose, reduce=reduce, prefix=prefix)


def dispatch(
        train: float, test: float, dev: float, config: str, output_dir: str,
        verbose=False, concat: bool = False) -> ProtogenieConfiguration:
    """

    :param train:
    :param test:
    :param dev:
    :param config:
    :param output_dir:
    :param verbose:
    :param concat:
    :return: PPAConfiguration for test purposes
    """

    train, test, dev = check_ratio(train, test, dev)
    config = ProtogenieConfiguration.from_xml(config)
    no_split = sorted([train, test, dev]) == [0, 0, 1]

    os.makedirs(output_dir, exist_ok=True)
    if not no_split:  # No split
        for subset in ["dev", "test", "train"]:
            os.makedirs(os.path.join(output_dir, subset), exist_ok=True)

    print("=============")
    print("Processing...")
    # I run over each files
    for file, ratios in split_files(output_folder=output_dir, verbose=verbose, dev_ratio=dev, test_ratio=test,
                                    config=config, no_split=no_split):

        print("{} has been transformed".format(file))
        for key, value in ratios.items():
            if value:
                print("\t{} tokens in {} dataset".format(value, key))

    return config


def from_memory(memory_file: str, config: str, output_dir: str,
                dev_ratio: float = None, test_ratio: float = None) -> ProtogenieConfiguration:
    config = ProtogenieConfiguration.from_xml(config)

    os.makedirs(output_dir, exist_ok=True)
    for subset in ["dev", "test", "train"]:
        os.makedirs(os.path.join(output_dir, subset), exist_ok=True)

    for file, ratios in files_from_memory(config=config, memory_file=memory_file, output_folder=output_dir,
                                          dev_ratio=dev_ratio, test_ratio=test_ratio):
        print("{} has been transformed".format(file))
        for key, value in ratios.items():
            if value:
                print("\t{} tokens in {} dataset".format(value, key))

    return config


def concat(config: str, output_dir: str, verbose: bool = True, reduce: Optional[float] = None,
           prefix: str = ""
           ) -> ProtogenieConfiguration:
    if reduce is not None:
        if reduce > 1 or reduce < 0.:
            raise click.BadParameter("Reduce should be in the range ]0:1]")
    config = ProtogenieConfiguration.from_xml(config)
    dataset = None

    max_len = 50
    template = '    {:'+str(max_len)+'s} {:>10d} {:>10d}'
    chunks_lines = {}
    for data_type, filename, nb_chunks, nb_lines in glue(
            config=config, output_folder=output_dir, verbose=verbose, reduce=reduce, prefix=prefix):
        if dataset != data_type:
            if dataset in chunks_lines:
                click.echo("# {}'s statistics".format(dataset))
                click.echo("Chunks: {:>15,d}".format(chunks_lines[dataset][0]))
                click.echo("Tokens: {:>15,d}".format(chunks_lines[dataset][1]))
            if verbose:
                click.echo("\n========\n{}\n".format(data_type))
                click.echo(template.replace("d", "s").format("File", "Chunks", "Tokens"))
            dataset = data_type
            chunks_lines[data_type] = [0, 0]
        if verbose:
            click.echo(template.format(filename[:max_len], nb_chunks, nb_lines))
        chunks_lines[data_type][0] += nb_chunks
        chunks_lines[data_type][1] += nb_lines

    click.echo("# {}'s statistics".format(dataset))
    click.echo("Chunks: {:>15,d}".format(chunks_lines[dataset][0]))
    click.echo("Tokens: {:>15,d}".format(chunks_lines[dataset][1]))

    total_chunks, total_lines = sum([v[0] for v in chunks_lines.values() if v[0]]), \
                                sum([v[1] for v in chunks_lines.values() if v[1]])

    click.echo("++++++++++++++\nSummary:")
    for dataset in chunks_lines:
        click.echo("{}: {:.2f} of chunks, {:.2f} of tokens".format(
            dataset,
            chunks_lines[dataset][0] / total_chunks * 100,
            chunks_lines[dataset][1] / total_lines * 100
        ))
    return config



from .io_utils import add_sentence, get_name
from .configs import CorpusConfiguration, ProtogenieConfiguration
from dataclasses import dataclass
from typing import Dict, Optional, List, Set, Tuple
from .splitters import LineSplitter
from .reader import ColumnNotFound
import glob
import os
import math
import csv


__all__ = ["split_files", "files_from_memory", "ConfigError"]


class ConfigError(Exception):
    """ Error when a configuration has been wrong"""


@dataclass
class _Range:
    """Just a class to deal with typing. End is the end of the range, dataset is [train|dev|test]"""
    end: int
    dataset: str


@dataclass
class _CorpusDispatched:
    """ Item that contains informations about dispatching

    Using dataclass mainly for typing"""
    config: CorpusConfiguration
    lines: Dict[int, _Range]


################
# Main Functions
################


def split_files(
        config: ProtogenieConfiguration, output_folder: str, dev_ratio: float, test_ratio: float,
        verbose: bool = True, no_split: bool = False):
    """ Dispatch sentence for each file in files

    :param config: Configuration for PPA Splitter
    :param output_folder: Folder where the data should be saved
    :param dev_ratio: Ratio of data to put in dev
    :param test_ratio: Ratio of data to put in test
    :param verbose: Verbosity (Adds some print during process)
    :param no_split: Do not apply splitting

    :yield: File, Dispatch stats about file
    """

    memory, memory_file = None, None
    if config.memory:
        memory_file = open(config.memory, "w")
        memory = csv.writer(memory_file)

    # For each file
    for unix_path, current_config in config.corpora.items():
        unix_path = os.path.join(config.dir, unix_path)
        for file in glob.glob(unix_path):
           yield from _single_file_dispatch(
               file, current_config=current_config, memory=memory,
               dev_ratio=dev_ratio, test_ratio=test_ratio, output_folder=output_folder,
               config=config, verbose=verbose, no_split=no_split
           )

    if memory:
        memory_file.close()


def files_from_memory(
        config: ProtogenieConfiguration, output_folder: str, memory_file: str,
        verbose: bool = True, dev_ratio: float = None, test_ratio: float = None):
    """ Regenerate a corpus using the same previously selected lined but potentially
    adding files and different post-processing

    :param config: Configuration
    :param output_folder: Directory where to save files
    :param memory_file: Memory file that holds lines to dispatch
    :param verbose: Whether to print stuff
    :param dev_ratio: Dev Ratio
    :param test_ratio: Test ratio
    """
    with open(memory_file) as memory:
        memory_reader = list(csv.reader(memory))

    dispatcher: Dict[str, _CorpusDispatched] = {
        os.path.realpath(real_path): _CorpusDispatched(config=corpus_config, lines={})
        for unix_path, corpus_config in config.corpora.items()
        for real_path in glob.glob(os.path.join(config.dir, unix_path))
    }

    # For each file, we build a map of the lines that needs to be dispatched
    for line in memory_reader:
        if not line:
            continue
        current_file, line_range, dataset_target = line

        real_path = os.path.realpath(current_file)
        if real_path in dispatcher:
            start, end = tuple(map(int, line_range.split("-")))
            dispatcher[real_path].lines[start] = _Range(end=end, dataset=dataset_target)

    new_files = []
    for file, dispatching in dispatcher.items():
        if not dispatching.lines:
            new_files.append((file, dispatching.config))
            pass

        # We set up a dictionary of token count to print nice
        #  information later
        training_tokens = {"test": 0, "dev": 0, "train": 0}

        current_config = dispatching.config

        header_line = []
        created_files = set()
        sentence = []
        blanks = 0
        current_set: Optional[_Range] = None

        with open(file) as f:
            for line_no, line in enumerate(f):
                if line_no == 0:
                    if current_config.reader.has_header:
                        header_line = current_config.reader.set_header(line)
                        continue
                    else:
                        header_line = current_config.reader.header
                elif not line.strip() and not isinstance(current_config.splitter, LineSplitter):
                    # Only count is we already have written or the sentence writing has started
                    if len(sentence) > 0:
                        blanks += 1
                    continue

                if line_no in dispatching.lines:                  # We begin a set
                    current_set = dispatching.lines[line_no]
                    sentence.append(line)
                elif current_set and line_no != current_set.end:  # We are in the set
                    sentence.append(line)
                elif current_set and line_no == current_set.end:  # We are at the end of the set
                    sentence.append(line)
                    blanks = 0  # ToDo: Blanks are not really taken into account here...
                    sentence = [x for x in sentence if x.strip()]
                    add_sentence(
                        output_folder=output_folder,
                        dataset=current_set.dataset,
                        filename=file,
                        sentence=sentence,
                        source_marker=current_config.column_marker,
                        output_marker=config.output.column_marker
                    )
                    training_tokens[current_set.dataset] += len(sentence)
                    sentence = []

            # Finally, if there is something remaining
            if len(sentence) and current_set:
                add_sentence(
                    output_folder=output_folder,
                    dataset=current_set.dataset,
                    filename=file,
                    sentence=sentence,
                    source_marker=current_config.column_marker,
                    output_marker=config.output.column_marker
                )
                training_tokens[current_set.dataset] += len(sentence)

        # Add the header to the files
        created_files.update(
            _add_header(
                output_folder=output_folder, training_tokens=training_tokens, header_line=header_line,
                current_config=current_config, file=file
            )
        )

        yield file, training_tokens

        if config.postprocessings:
            for post_processings in config.postprocessings:
                for output_file in created_files:
                    post_processings.apply(output_file, current_config)

    memory = open(memory_file, "w")
    writer = csv.writer(memory)
    writer.writerows(memory_reader)

    if new_files:
        # We have new files, we need to deal with them per usual
        if not test_ratio:
            raise ConfigError("Ratios were not given and we have a new file.")

        for file, current_config in new_files:
            yield from _single_file_dispatch(
                config=config,
                dev_ratio=dev_ratio,
                test_ratio=test_ratio,
                current_config=current_config,
                verbose=verbose,
                file=file,
                output_folder=output_folder,
                memory=writer
            )
    memory.close()


def glue(config: ProtogenieConfiguration, output_folder: str,
         verbose: bool = True, reduce: Optional[float] = None, prefix: str = ""):
    """

    :param config:
    :param output_folder:
    :param verbose:
    :param reduce: [Optional] Take only a portion of a float
    :return: Generator[data_type, filename, nb_chunks, nb_lines]
    """
    def write_sentence(sents: List[List[str]], dataset):
        dataset.write("\n".join(config.output.column_marker.join(sent) for sent in sents)+"\n\n") # 2 new lines

    for dataset_type in ["train", "test", "dev"]:
        cur_dir = os.path.join(output_folder, dataset_type)

        out = os.path.join(output_folder, prefix + dataset_type + ".tsv")
        out = open(out, "w")
        out.write(config.output.column_marker.join(config.output.header)+"\n")

        for _, _, files in os.walk(cur_dir, topdown=False):
            for file in files:
                basename = os.path.basename(file)
                tokens = [0]

                # Output files necessarly have headers and empty lines as markers
                with open(os.path.join(cur_dir, file)) as f:
                    sentences = []
                    sentence = []
                    for line_numb, line in enumerate(f.readlines()):
                        line = line.strip()
                        # If we met the header
                        if line_numb == 0:
                            cols = line.split(config.output.column_marker)
                            header_map = [cols.index(head) for head in config.output.header]
                            continue

                        # If we have a sentence
                        if not line and sentence:
                            sentences.append(sentence)
                            sentence = []
                            tokens.append(0)
                            continue

                        mapped = line.split(config.output.column_marker)
                        if not "".join(mapped).strip():
                            continue
                        try:
                            mapped = [mapped[index] for index in header_map]
                        except IndexError:
                            raise ValueError(f"Line {line_numb} failing in {file}: {mapped}")
                        tokens[-1] += 1
                        sentence.append(mapped)

                    if sentence:
                        sentences.append(sentence)

                    max_int = len(sentences)
                    if reduce:# and dataset_type == "train":
                        max_int = math.ceil(reduce * max_int)

                    for sentence in sentences[:max_int]:
                        #if dataset_type == "dev":
                        #    print(sentence)
                        write_sentence(sentence, out)

                yield dataset_type, basename, len(sentences[:max_int]), sum(tokens[:max_int])

        out.close()
    return []

##################
# Shared functions
##################


def _preview(file: str, current_config: CorpusConfiguration) -> Tuple[List[str], int, int, int]:
    # We count things in the file
    unit_counts = 0
    empty_lines = 0
    header_line = []

    if not current_config.reader.has_header:
        header_line = current_config.reader.header

    with open(file) as f:
        for line_no, line in enumerate(f):
            if line_no == 0:
                if current_config.reader.has_header:
                    header_line = current_config.reader.set_header(line)
                    continue

            if line_no == 0 and current_config.reader.has_header:
                continue  # Skip the first line in count if we have a header
            try:
                unit_counts += int(current_config.splitter(line, reader=current_config.reader))
            except ColumnNotFound as E:
                print(f"ERROR: Line {line_no} is badly formated, column not found error encountered. "
                      f"Text=`{line.strip()}`")
            empty_lines += int(not bool(line.strip()))  # Count only lines if they are empty

    return header_line, unit_counts, empty_lines, line_no - int(current_config.reader.has_header) - empty_lines + 1


def _single_file_dispatch(
        file: str, current_config: CorpusConfiguration,
        config: ProtogenieConfiguration, output_folder: str, dev_ratio: float, test_ratio: float,
        verbose: bool = True, memory=None, no_split: bool = False
):
    # We do two passes here
    #  1. The first one is used to collect informations about the file. In order to not keep data in memory,
    #     we iterate over it and count the number of real lines + the number of sentences.
    #     Sentences are counted on the base of the Configuration.split function
    #  2. We read the file again and dispatch according to the ratio and the data we got before
    #      Note : We use .pop(0) to move from start to end. If we have one day a performance issue
    #      we might want to move to a yield system
    #
    # This method is slower but allows for memory efficiency.

    header_line, unit_counts, empty_lines, lines = _preview(file, current_config)

    if verbose:
        if no_split is True:
            print("Not splitting file {filename}".format(filename=file))
        else:
            print("{unit_count} {unit_name} to dispatch in {filename} ({lines} full, {lines_empty} empty)".format(
                filename=file, unit_name=current_config.unit_name, unit_count=unit_counts,
                lines_empty=empty_lines, lines=lines
            ))

    # We set up numbers based on the ratio
    # In order to do that, we get to use
    if no_split is False:
        target_dataset = current_config.build_dataset_dispatch_list(
            units_count=unit_counts,
            test_ratio=test_ratio,
            dev_ratio=dev_ratio
        )

        # We set up a dictionary of token count to print nice
        #  information later
        training_tokens = {"test": 0, "dev": 0, "train": 0}
    else:
        training_tokens = {"output": 0}

    # ToDo: When file splitter, the number of lines should be passed here probably ? Or is reset the issue ? ...

    current_config.splitter.reset()
    if not no_split:
        current_config.splitter.set_targets(target_dataset)

    created_files = set()

    with open(file) as f:
        sentence = []
        blanks = 0
        for line_no, line in enumerate(f):
            if line_no == 0 and current_config.reader.has_header:
                continue
            elif not line.strip() and not isinstance(current_config.splitter, LineSplitter):
                # Only count is we already have written or the sentence writing has started
                if len(sentence) > 0:
                    blanks += 1
                continue

            try:
                is_a_split = current_config.splitter(line, reader=current_config.reader)
            except ColumnNotFound:
                print(f"ERROR: Line {line_no} is badly formated, column not found error encountered. "
                      f"Text=`{line.strip()}`")
                continue

            sentence.append(line)
            if is_a_split:
                if no_split:
                    dataset = "output"
                else:
                    dataset = target_dataset.pop(0)

                if memory:
                    memory.writerow([os.path.relpath(file),
                                     "{}-{}".format(line_no - len(sentence) + 1 - blanks, line_no),
                                     dataset])
                    blanks = 0

                sentence = [x for x in sentence if x.strip()]
                add_sentence(
                    output_folder=output_folder,
                    dataset=dataset,
                    filename=file,
                    sentence=sentence,
                    source_marker=current_config.column_marker,
                    output_marker=config.output.column_marker,
                    subfolder=not no_split
                )
                training_tokens[dataset] += len(sentence)
                sentence = []

        # Finally, if there is something remaining
        if len(sentence):
            try:
                dataset = target_dataset.pop(0)
                # print("last dataset ?")
            except Exception:
                if no_split:
                    dataset = "output"
                else:
                    dataset = "train"

            if memory:
                memory.writerow([file, "{}-{}".format(line_no - len(sentence) + 1 - blanks, line_no), dataset])

            add_sentence(
                output_folder=output_folder,
                dataset=dataset,
                filename=file,
                sentence=sentence,
                source_marker=current_config.column_marker,
                output_marker=config.output.column_marker,
                subfolder=not no_split
            )
            training_tokens[dataset] += len(sentence)

    created_files.update(
        _add_header(
            output_folder=output_folder, training_tokens=training_tokens, header_line=header_line,
            current_config=current_config, file=file,
            subfolder=not no_split
        )
    )

    yield file, training_tokens

    if config.postprocessings:
        for post_processings in config.postprocessings:
            for output_file in created_files:
                post_processings.apply(output_file, current_config)


def _add_header(output_folder: str, file: str,
                training_tokens: Dict[str, int], current_config: CorpusConfiguration,
                header_line: List[str],
                subfolder: bool = True
                ) -> Set[str]:
    files = set()
    for dataset, tokens in training_tokens.items():
        if tokens:
            if subfolder:
                trg = get_name(output_folder, dataset, file)
            else:
                trg = get_name(output_folder, "", file)
            files.add(trg)  # We add the file to the one we created
            with open(trg) as f:
                content = f.read()
            with open(trg, "w") as f:
                f.write(current_config.column_marker.join(header_line)+"\n"+content)
    return files



""" This module contains scripts that are supposed to be used to know
when the CSV reader enters a new sentence.

"""

import regex as re
import math
import random
from typing import Dict, Union, List, Tuple, Optional
if not True:
    from .configs import CorpusConfiguration
    from .reader import Reader

from .defaults import DEFAULT_SENTENCE_MARKERS


SPACE_ONLY = re.compile("^\s+$")


class _DispatcherSequential:
    @staticmethod
    def dispatch(units_count, test_ratio=0.2, dev_ratio=0.0001) -> List[str]:
        """ Get the ratios and builds a list of targets completely randomly and shuffled

        :param units_count: Number of units to dispatch (either sentence or dispatch depending on self.splitter)
        :param test_ratio: Ratio of data to be put in test
        :param dev_ratio: Ratio of data to be put in dev
        :return: List of dataset to dispatch to
        """

        train_number = units_count
        dev_number = 0
        if dev_ratio > 0.001:
            dev_number = int(math.ceil(dev_ratio * units_count))
            train_number = train_number - dev_number
        test_number = int(math.ceil(test_ratio * units_count))
        train_number = train_number - test_number

        target_dataset = ["train"] * train_number + ["test"] * test_number + ["dev"] * dev_number
        return target_dataset


class _DispatcherRandom(object):
    @staticmethod
    def dispatch(units_count, test_ratio=0.2, dev_ratio=0.0001) -> List[str]:
        """ Get the ratios and builds a list of targets completely randomly and shuffled

        :param units_count: Number of units to dispatch (either sentence or dispatch depending on self.splitter)
        :param test_ratio: Ratio of data to be put in test
        :param dev_ratio: Ratio of data to be put in dev
        :return: List of dataset to dispatch to
        """

        target_dataset = _DispatcherSequential.dispatch(units_count, test_ratio, dev_ratio)
        random.shuffle(target_dataset)
        return target_dataset


class _SplitterPrototype(_DispatcherRandom):
    """ This will contain any needed function accross splitters
    """
    def __init__(self, *args, **kwargs):
        pass

    def reset(self) -> None:
        """ Reset the splitter values for the second pass if necessary
        """

    def _repr_options(self) -> str:
        """ Return options of the splitter"""
        return ""

    def __repr__(self) -> str:
        return "<splitter name='{}'{}/>".format(self.__class__.__name__, self._repr_options())

    def read_line(self, line, corpus_configuration: "CorpusConfiguration") -> Dict[str, str]:
        line = line.split(corpus_configuration.column_marker)
        return line

    def __call__(self, line, reader: "Reader") -> bool:
        """
        """
        raise NotImplemented

    def set_targets(self, targets: List[str]) -> None:
        pass


class RegExpSplitter(_SplitterPrototype):
    def __init__(self, column_marker="\t", matchPattern: List[str] = None, source: List[str] = None, **kwargs):
        """ Returns true if the line is a sentence splitter by being empty

        :param column_marker: Marker that splits column in the CSV/TSV
        :param sentence_splitter: Marker that shows the end of a sentence

        >>> from protogenie.reader import Reader
        >>> reader = Reader(reader_type="explicit", keys=[("form", "token"), ("lemma", "lemma"), ("POS", "POS")],
        ...                 column_marker=" ")
        >>> reader.set_header("form lemma POS")
        ['token', 'lemma', 'POS']
        >>> obj = RegExpSplitter(" ", matchPattern=[r"PONfrt", r"Ref\."], source=["POS", "lemma"])
        >>> obj("abc abc PONfbl", reader)
        False
        >>> obj("abc Ref. OUT", reader)
        True
        >>> obj(". . PONfrt", reader)
        True

        """
        self.column_marker = column_marker
        self.match_pattern: List[str] = matchPattern or ["["+DEFAULT_SENTENCE_MARKERS+"]"]
        self.matcher: List[re.Regex] = [re.compile(matcher) for matcher in self.match_pattern]
        self.source: List[str] = source or ["form"]

    def _repr_options(self):
        return " matchPattern='{}'".format(self.match_pattern)

    def __call__(self, line: str, reader: "Reader" = None):
        for matcher, source in zip(self.matcher, self.source):
            if matcher.search(
                reader.get_column(
                    line.strip().split(reader.column_marker),
                    source,
                    raise_on_none=True
                )
            ):
                return True
        return False


class LineSplitter(_SplitterPrototype):
    def __init__(self, **kwargs):
        """ Class for applying a split on new lines (some data are formatted such as sentence are separated
        with empty lines
        """
        # We set it to True so that starting empty lines are
        #  not counting as separators
        self.last_line_was_empty = True

    def __call__(self, line, reader=None):
        if line == "\n":
            if self.last_line_was_empty:
                return False
            self.last_line_was_empty = True
            return True
        self.last_line_was_empty = False
        return False

    def reset(self):
        """ Reset the last line to True in case the file starts with empty lines
        """
        self.last_line_was_empty = True


class TokenWindowSplitter(_SplitterPrototype):
    def __init__(self, window=20, **kwargs):
        """ Class for applying a split every N words

        :param window: Split as a sentence each N words
        """
        self.window = int(window)
        self.words = 0

    def __call__(self, line, reader=None):
        if not line.strip():
            return False
        # No body cares about line in here
        self.words += 1
        if self.words == self.window:
            self.words = 0
            return True
        return False

    def _repr_options(self):
        return " window='{}'".format(self.window)

    def reset(self):
        self.words = 0


class FileSplitter(_SplitterPrototype):
    def __init__(self, *args, **kwargs):
        """ Splitter that affects the file globally. Not expecting random shuffle.
        The real main piece of this splitter is actually the line_counting function
        """
        super(FileSplitter, self).__init__(self, *args, **kwargs)
        self.line_count = 0
        self._sets_size: Dict[str, int] = {}
        self._dispatching: bool = False  #
        self._sizes: Optional[List[int]] = None
        
    def __call__(self, line, reader=None):
        if not line.strip():
            return False

        if not self._dispatching:  # If we are counting lines
            return True

        if self._sizes:
            self._sizes[0] -= 1
            # If we still have something in our size, we reduce the number
            if self._sizes[0]:
                return False
            else:
                self._sizes.pop(0)
                return True
        else:
            raise Exception("Sizes should be set and were not found")

    def reset(self) -> None:
        self.line_count = 0
        self._dispatching = True
        self._sizes = [
            self._sets_size.get("train", 0),
            self._sets_size.get("dev", 0),
            self._sets_size.get("test", 0),
        ]

    def dispatch(self, units_count, test_ratio=0.2, dev_ratio=0.0001) -> List[str]:
        """ Get the ratios and builds a list of targets completely randomly and shuffled

        :param units_count: Number of units to dispatch (either sentence or dispatch depending on self.splitter)
        :param test_ratio: Ratio of data to be put in test
        :param dev_ratio: Ratio of data to be put in dev
        :return: List of dataset to dispatch to
        """
        train_number = units_count
        dev_number = 0
        if dev_ratio > 0.01:
            dev_number = int(math.ceil(dev_ratio * units_count))
            train_number = train_number - dev_number
        test_number = int(math.ceil(test_ratio * units_count))
        train_number = train_number - test_number

        target_dataset = ["train"] * train_number + ["test"] * test_number + ["dev"] * dev_number
        self._sets_size = {
            key: target_dataset.count(key)
            for key in filter(lambda x: x in target_dataset, ["train", "dev", "test"])
        }
        return list(filter(lambda x: x in target_dataset, ["train", "dev", "test"]))



if False:
    from .configs import CorpusConfiguration
import tempfile
import regex as re
from xml.etree.ElementTree import Element
import csv
from typing import List
from .postprocessing import ApplyTo, PostProcessing


class RomanNumeral(PostProcessing):
    NodeName = "toolbox"

    def __init__(
            self, apply_to: ApplyTo, regex_change=None,
    ):
        """ Transform Roman Numerals to Integers

        :param apply_to: List of fields that should have the rules applied to
        """
        super(RomanNumeral, self).__init__()
        # Just thank you, dear anonymous hero named Corin
        # User: https://stackoverflow.com/users/302306/corin
        # Source: https://stackoverflow.com/a/10441405
        if regex_change:
            self.match_pattern: re.Regex = re.compile(regex_change)
        else:
            self.match_pattern: re.Regex = re.compile(r"^(M{1,4}(CM|CD|D?C{0,3})(XC|XL|L?X{0,3})(IX|IV|V?I{0,3})|M{0,4}"
                                                      r"(CM|C?D|D?C{1,3})(XC|XL|L?X{0,3})(IX|IV|V?I{0,3})|M{0,4}"
                                                      r"(CM|CD|D?C{0,3})(XC|X?L|L?X{1,3})(IX|IV|V?I{0,3})|M{0,4}"
                                                      r"(CM|CD|D?C{0,3})(XC|XL|L?X{0,3})(IX|I?V|V?I{1,3}))$")
        self.apply_to: ApplyTo = apply_to

    def apply(self, file_path: str, config: "CorpusConfiguration"):
        temp = tempfile.TemporaryFile(mode="w+")  # 2

        try:
            with open(file_path) as file:
                csv_reader = csv.reader(file, delimiter=config.column_marker)
                header: List[str] = []
                for nb_line, line in enumerate(csv_reader):  # The file should already have been open
                    if nb_line == 0:
                        temp.write(config.column_marker.join(line)+"\n")
                        header = line
                        continue
                    elif not line:
                        temp.write("\n")
                        continue
                    lines = dict(zip(header, line))

                    if self.match_pattern.search(lines[self.apply_to.source]):
                        # original = {k: v for k, v in lines.items()}
                        result = str(self.from_roman(lines[self.apply_to.source]))

                        for target in self.apply_to.target:
                            # If source and target are the same, we simply replace source by target
                            if self.apply_to.source == target:
                                lines[self.apply_to.source] = result
                            else:  # Otherwise, we just copy the result value to the target
                                lines[target] = result

                    temp.write(config.column_marker.join(list(lines.values()))+"\n")
            with open(file_path, "w") as f:
                temp.seek(0)
                f.write(temp.read())
        finally:
            temp.close()  # 5

    @classmethod
    def from_xml(cls, node: Element) -> "RomanNumeral":
        return cls(
            apply_to=ApplyTo.from_xml(node.find("applyTo")), regex_change=node.get("matchPattern", None)
        )

    @staticmethod
    def from_roman(num: str) -> int:
        """

        Source: https://stackoverflow.com/questions/19308177/converting-roman-numerals-to-integers-in-python
        Author: https://stackoverflow.com/users/1201737/r366y
        :param num:
        :return:
        """
        roman_numerals = {'I': 1, 'V': 5, 'X': 10, 'L': 50, 'C': 100, 'D': 500, 'M': 1000}
        result = 0
        for i,c in enumerate(num.upper()):
            if (i+1) == len(num) or roman_numerals[c] >= roman_numerals[num[i+1]]:
                result += roman_numerals[c]
            else:
                result -= roman_numerals[c]
        return result

    @classmethod
    def match_config_node(cls, node: Element) -> bool:
        return node.tag == cls.NodeName and node.get("name") == "RomanNumeral"



from typing import Dict, Optional, Any, Type, List
import os.path
import lxml.etree as ET
from copy import deepcopy
from collections import defaultdict

from .splitters import RegExpSplitter, LineSplitter, TokenWindowSplitter, FileSplitter, _SplitterPrototype
from .reader import Reader
from .postprocessing import Disambiguation, ReplacementSet, Skip, PostProcessing, Clitic, Capitalize
from .toolbox import RomanNumeral
import datetime
from dataclasses import dataclass
Splitter = Type[_SplitterPrototype]

PostProcessingClasses = [Disambiguation, ReplacementSet, Skip, RomanNumeral, Clitic, Capitalize]


@dataclass
class Output:
    """ Details about output"""
    header: List[str]
    column_marker: str
    readable_column_marker: str


class CorpusConfiguration:
    SPLITTERS = {
        "empty_line": LineSplitter,
        "regexp": RegExpSplitter,
        "token_window": TokenWindowSplitter,
        "file_split": FileSplitter
    }

    UNIT_NAMES = {
        "empty_line": "chunks",
        "regexp": "sentences",
        "token_window": "token-windows",
        "file_split": "tokens"
    }

    COLUMN_SPECIAL_MARKERS = {
        "TAB": "\t"
    }
    REVERSE_COLUMN_SPECIAL_MARKERS = {
        character: name
        for name, character in COLUMN_SPECIAL_MARKERS.items()
    }

    def __init__(self,
                 reader: Reader,
                 splitter: str = "sentence_marker", **spliter_options):
        """ Initiate a configuration which will allow file-basis dispatching

        :param splitter: The splitter to use (Default : sentence_marker)
        :param sentence_markers: Characters to use for separating on sentence marker
        :param each_n_words:
        :param column_marker: Marker for columns in files
        """
        self.reader: Reader = reader
        self.splitter_name: str = splitter

        _spliter_options: Dict[str, Any] = spliter_options

        # Some characters are awful to replicate in YAML, that's why
        #  we have a dictionary of simplified names
        _spliter_options["column_marker"] = self.COLUMN_SPECIAL_MARKERS.get(
            _spliter_options["column_marker"],
            _spliter_options["column_marker"]
        )

        self.column_marker: str = _spliter_options["column_marker"]

        # We set up the splitter
        splitter_class: Optional[Type] = self.SPLITTERS.get(self.splitter_name, None)
        if splitter_class is None:
            raise ValueError("Splitter '{}' is not in the acceptable list: {}".format(
                self.splitter_name, ", ".join(list(self.SPLITTERS.keys()))
            ))
        # Initialize the splitter
        self.splitter: _SplitterPrototype = splitter_class(**_spliter_options)

    def __repr__(self):
        return "<corpus column_marker='{column}'>{splitter}</corpus>".format(
            column=self.column_marker.replace("\t", "TAB"),
            splitter=self.splitter
        )

    @property
    def unit_name(self):
        return self.UNIT_NAMES[self.splitter_name]

    def splitter_reset(self):
        """ This function can be useful to reset a splitter in between passes
        """
        self.splitter.reset()

    def build_dataset_dispatch_list(self, units_count, test_ratio=0.2, dev_ratio=0.0001):
        """ Build a list of dataset target that will be used by the dispatching loop


        :param units_count: Number of units to dispatch (either sentence or dispatch depending on self.splitter)
        :param test_ratio: Ratio of data to be put in test
        :param dev_ratio: Ratio of data to be put in dev
        :return: List of dataset name targets (list of "train", "test" and "dev")
        """
        return self.splitter.dispatch(units_count, test_ratio=test_ratio, dev_ratio=dev_ratio)


class ProtogenieConfiguration:
    def __init__(self,
                 path: str,
                 corpora: Dict[str, CorpusConfiguration],
                 output: Output,
                 memory: Optional[str] = None,
                 postprocessings: List[PostProcessing] = None,
                 **kwargs
                 ):
        self.path: str = path
        self.dir: str = os.path.abspath(os.path.dirname(path))

        self.memory: Optional[str] = memory
        self.corpora: Dict[str, CorpusConfiguration] = corpora
        self.postprocessings: List[PostProcessing] = postprocessings or []
        self.output: Output = output

    @staticmethod
    def read_splitter_params(node: ET.Element):
        """ Parses a node and provides a dictionary of parameters

        >>> elem = ET.fromstring("<splitter name='test' uppercase='true'><option a='1' b='2'/><option a='3' /></splitter>")
        >>> ProtogenieConfiguration.read_splitter_params(elem)
        {'uppercase': 'true', 'a': ['1', '3'], 'b': ['2']}
        """
        base = {key: value for key, value in node.attrib.items() if key != "name"}
        options = node.xpath("./option")
        if options:
            adding = defaultdict(list)
            for opt in options:
                for key, value in opt.attrib.items():
                    adding[key].append(value)
            base.update(adding)
        return base

    @classmethod
    def from_xml(cls, filepath: str) -> "ProtogenieConfiguration":
        with open(filepath) as f:
            xml = ET.parse(f)
            inc = ET.XInclude()
            inc(xml.getroot())
        kwargs = {}

        # Get readers
        default_reader = Reader.from_xml(xml.find("./default-header/header"), default=None)

        col = xml.find("./output").attrib["column_marker"]
        header_node = xml.find("./output/header")
        if not header_node or header_node.get("name", "default") == "default":
            header = [mapped or key for key, mapped in default_reader._keys]
        else:
            header = [node.text.strip() for node in xml.findall("./output/header/key")]
        output = Output(
            column_marker=col.replace("TAB", "\t"),
            readable_column_marker=col,
            header=header
        )

        # Parse corpora configurations
        corpora = {}
        for corpus in xml.findall("./corpora/corpus"):
            corpora[corpus.get("path")] = CorpusConfiguration(**{
                "column_marker": corpus.get("column_marker"),
                "splitter": corpus.find("splitter").get("name"),
                "reader": Reader.from_xml(corpus.find("./header"), default=default_reader),
                **cls.read_splitter_params(corpus.find("./splitter"))
            })

        # Check options
        if len(xml.findall("./memory")):
            kwargs["memory"] = xml.find("./memory").get("path")

            kwargs["memory"] = kwargs["memory"].replace(
                "$file$", os.path.splitext(os.path.basename(filepath))[0]
            ).replace(
                "$date$", datetime.datetime.now().strftime("%y-%m-%d_%H-%M-%S")
            )

        kwargs["postprocessings"] = []
        for element in xml.findall("./postprocessing"):
            for child in element:
                for post_processing_class in PostProcessingClasses:
                    if post_processing_class.match_config_node(child):
                        kwargs["postprocessings"].append(post_processing_class.from_xml(child))

        return cls(path=filepath, corpora=corpora, output=output, **kwargs)



import tempfile
import math
import random
from abc import ABC, abstractmethod
from collections import namedtuple
from xml.etree.ElementTree import Element
import csv
from typing import List, ClassVar, Tuple, Dict, Optional, TYPE_CHECKING, Union

import regex as re

if TYPE_CHECKING:
    from .configs import CorpusConfiguration
from .sentence_matchers import SentenceMatcherProto, SentenceRegexpMatcher
Numeric = Union[int, float]


class PostProcessing(ABC):
    NodeName = "XML-NODE-LOCAL-NAME"  # Name of the node to match

    @abstractmethod
    def apply(self, file_path: str, config: "CorpusConfiguration"):
        raise NotImplementedError

    @abstractmethod
    def from_xml(cls, node: Element) -> ClassVar["PostProcessing"]:
        raise NotImplementedError

    @classmethod
    def match_config_node(cls, node: Element) -> bool:
        """ If the current node is representing the current object, returns True
        """
        return node.tag == cls.NodeName

    def _modify_line(self, header: List[str], values: Optional[List[str]],
                     file_path: str, config: "CorpusConfiguration"):
        raise NotImplementedError

    def _stop_chunk(self, line: Optional[Dict[str, str]]) -> bool:
        raise NotImplementedError

    def _chunk_modify_routine(self, file_path: str, config: "CorpusConfiguration"):
        raise NotImplementedError

    def _scan_chunks(self,
                     file_path: str, config: "CorpusConfiguration",
                     sentence_matcher: Optional[SentenceMatcherProto]) -> Tuple[int, int]:
        """ Analyzes the FILE for the number of chunks

        """
        chunks = 0
        tokens = 0
        with open(file_path) as file:
            for nb_line, line in enumerate(file):
                vals = line.strip().split(config.column_marker)

                if nb_line == 0:
                    header = vals
                    continue

                if len(header) == len(vals):
                    tokens += 1
                    if sentence_matcher and sentence_matcher.match(header, vals):
                        chunks += 1
                elif sentence_matcher:
                    chunks += sentence_matcher.match(header, None)

        return chunks, tokens

    def _single_line_modify_routine(self, file_path: str, config: "CorpusConfiguration"):
        header: List[str] = []
        temp = tempfile.TemporaryFile(mode="w+")  # 2

        try:
            with open(file_path) as file:
                for nb_line, line in enumerate(file):

                    if not line.strip():
                        temp.write(line)
                        self._modify_line(header, None, file_path, config)
                        continue

                    vals = line.strip().split(config.column_marker)

                    if nb_line == 0:
                        header = vals
                        temp.write(line)
                        continue

                    modified = self._modify_line(header, vals, file_path=file_path, config=config)
                    temp.write(
                        config.column_marker.join(
                            [modified[head] for head in header]
                        ) + "\n"
                    )

            with open(file_path, "w") as f:
                temp.seek(0)
                f.write(temp.read())
        finally:
            temp.close()  # 5


class ApplyTo:
    def __init__(self, source: str, target: List[str]):
        self.source: str = source
        self.target: List[str] = target

    @staticmethod
    def from_xml(apply_to_node: Element) -> "ApplyTo":
        return ApplyTo(
            source=apply_to_node.attrib["source"],
            target=[str(node.text).strip() for node in apply_to_node.findall("./target")]
        )


class Disambiguation(PostProcessing):
    NodeName = "disambiguation"

    def __init__(self, lemma_key: str, disambiguation_key: str, match_pattern: str,
                 default_value: str, glue: str, keep: bool = False):
        super(Disambiguation, self).__init__()
        self.lemma_key: str = lemma_key
        self.disambiguation_key: str = disambiguation_key
        self.match_pattern: re.Regex = re.compile(match_pattern)
        self.keep: bool = keep
        self.default_value: str = default_value
        self.glue: str = glue

    def apply(self, file_path: str, config: "CorpusConfiguration"):
        temp = tempfile.TemporaryFile(mode="w+")  # 2

        try:
            with open(file_path) as file:
                csv_reader = csv.reader(file, delimiter=config.column_marker)
                header: List[str] = []
                for nb_line, line in enumerate(csv_reader):  # The file should already have been open
                    if nb_line == 0:
                        temp.write(config.column_marker.join(line + [self.disambiguation_key]) + "\n")
                        header = line
                        continue
                    elif not line:
                        temp.write("\n")
                        continue
                    lines = dict(zip(header, line))

                    try:
                        found = self.match_pattern.findall(lines[self.lemma_key])
                    except KeyError:
                        print(lines, nb_line)
                        raise

                    if found:
                        lines[self.disambiguation_key] = found[0]
                        if not isinstance(found[0], str):
                            lines[self.disambiguation_key] = self.glue.join(found[0])
                        if not self.keep:  # If we do not keep the original value, we remove it
                            lines[self.lemma_key] = self.match_pattern.sub("", lines[self.lemma_key])
                    else:
                        lines[self.disambiguation_key] = self.default_value
                    temp.write(config.column_marker.join(list(lines.values())) + "\n")
            with open(file_path, "w") as f:
                temp.seek(0)
                f.write(temp.read())
        finally:
            temp.close()  # 5

    @classmethod
    def from_xml(cls, node: Element) -> "Disambiguation":
        return cls(
            lemma_key=node.attrib["source"],
            disambiguation_key=node.attrib["new-column"],
            match_pattern=node.attrib["matchPattern"],
            keep="keep" in node.attrib,
            default_value=node.attrib.get("default", "_"),
            glue=node.attrib.get("join", "|")
        )


class ReplacementSet(PostProcessing):
    """ Using a regular expression, replaces values in certain columns
    """
    NodeName = "replacement"

    def __init__(
            self, match_pattern: str, replacement_pattern: str,
            applies_to: List[ApplyTo]
    ):
        super(ReplacementSet, self).__init__()
        self.match_pattern: re.Regex = re.compile(match_pattern)
        self.replacement_pattern: str = replacement_pattern
        self.applies_to: List[ApplyTo] = applies_to

    def apply(self, file_path: str, config: "CorpusConfiguration"):
        temp = tempfile.TemporaryFile(mode="w+")
        try:
            with open(file_path) as file:
                csv_reader = csv.reader(file, delimiter=config.column_marker)
                header: List[str] = []
                for nb_line, line in enumerate(csv_reader):  # The file should already have been open
                    if nb_line == 0:
                        temp.write(config.column_marker.join(line) + "\n")
                        header = line
                        continue
                    elif not line:
                        temp.write("\n")
                        continue
                    lines = dict(zip(header, line))

                    for apply_to in self.applies_to:
                        if self.match_pattern.search(lines[apply_to.source]):
                            for target in apply_to.target:
                                # If source and target are the same, we simply replace source by target
                                if apply_to.source == target:
                                    lines[apply_to.source] = self.match_pattern.sub(
                                        self.replacement_pattern,
                                        lines[apply_to.source]
                                    )
                                else:  # Otherwise, we just set the target value using this value
                                    lines[target] = self.replacement_pattern

                    temp.write(config.column_marker.join(list(lines.values())) + "\n")
            with open(file_path, "w") as f:
                temp.seek(0)
                f.write(temp.read())
        finally:
            temp.close()  # 5

    @classmethod
    def from_xml(cls, node: Element) -> "ReplacementSet":
        return ReplacementSet(
            match_pattern=node.attrib["matchPattern"],
            replacement_pattern=node.attrib["replacementPattern"],
            applies_to=[ApplyTo.from_xml(apply_to) for apply_to in node.findall("applyTo")]
        )


class Skip(PostProcessing):
    """ If the matchPattern matches target column, the line is removed from the post-processed output
    """
    NodeName = "skip"

    def __init__(
            self, match_pattern: str, source: str
    ):
        super(Skip, self).__init__()
        self.match_pattern: re.Regex = re.compile(match_pattern)
        self.source: str = source

    def apply(self, file_path: str, config: "CorpusConfiguration"):
        temp = tempfile.TemporaryFile(mode="w+")  # 2

        try:
            with open(file_path) as file:
                csv_reader = csv.reader(file, delimiter=config.column_marker)
                header: List[str] = []
                for nb_line, line in enumerate(csv_reader):  # The file should already have been open
                    if nb_line == 0:
                        temp.write(config.column_marker.join(line) + "\n")
                        header = line
                        continue
                    elif not line:
                        temp.write("\n")
                        continue

                    lines = dict(zip(header, line))

                    # If it matches, we skip it
                    if self.match_pattern.search(lines[self.source]):
                        continue

                    temp.write(config.column_marker.join(list(lines.values())) + "\n")

            with open(file_path, "w") as f:
                temp.seek(0)
                f.write(temp.read())
        finally:
            temp.close()  # 5

    @classmethod
    def from_xml(cls, node: Element) -> "Skip":
        return Skip(
            match_pattern=node.attrib["matchPattern"],
            source=node.attrib["source"]
        )


class Clitic(PostProcessing):
    """ If the matchPattern matches target column, the line is removed from the post-processed output
    """
    NodeName = "clitic"

    Transfer = namedtuple("Transfer", ["col", "glue"])

    def __init__(
            self, match_pattern: str, source: str, glue: str, transfers: List[Tuple[str, bool]]
    ):
        super(Clitic, self).__init__()
        self.match_pattern: re.Regex = re.compile(match_pattern)
        self.source: str = source
        self.glue = glue
        _tr = {False: "", True: self.glue}
        self.transfers: List[Clitic.Transfer] = [
            Clitic.Transfer(key, _tr[has_glue])
            for key, has_glue in transfers
            if not print(key, has_glue)
        ]

    def apply(self, file_path: str, config: "CorpusConfiguration"):
        temp = tempfile.TemporaryFile(mode="w+")  # 2
        default = ("", "")
        try:
            with open(file_path) as file:
                csv_reader = csv.reader(file, delimiter=config.column_marker)
                header: List[str] = []
                sequence = []
                # [Int = Line to apply modifications to, Dict[Column name, Tuple[Glue, Value]]]
                modifications: List[Tuple[int, Dict[str, Tuple[str, str]]]] = []
                for nb_line, line in enumerate(csv_reader):  # The file should already have been open
                    if nb_line == 0:
                        temp.write(config.column_marker.join(line) + "\n")
                        header = line
                        continue
                    elif not line:
                        for target_line, modif in modifications:
                            sequence[target_line] = {
                                column: modif.get(column, default)[0].join(
                                    [value, modif.get(column, default)[1]]
                                )
                                for column, value in sequence[target_line].items()
                            }
                        temp.write("\n".join([
                            config.column_marker.join(list(l.values()))
                            for l in sequence
                        ]) + "\n")
                        sequence = []
                        modifications = []
                        continue

                    lines = dict(zip(header, line))

                    # If it matches, we give it to the previous / original line
                    if self.match_pattern.match(lines[self.source]):
                        modifications.append(
                            (
                                len(sequence) - 1 - len(modifications),
                                {key: (keep, lines[key]) for (key, keep) in self.transfers}
                            )
                        )
                        continue

                    # config.column_marker.join(list(lines.values()))
                    sequence.append(lines)

            with open(file_path, "w") as f:
                temp.seek(0)
                f.write(temp.read())
        finally:
            temp.close()  # 5

    @classmethod
    def from_xml(cls, node: Element) -> "Clitic":
        return cls(
            match_pattern=node.attrib["matchPattern"],
            source=node.attrib["source"],
            glue=node.attrib["glue_char"],
            transfers=[
                (
                    tr.text,
                    tr.attrib.get("no-glue-char", "false").lower() == "false"
                )
                for tr in node.findall("transfer")
            ]
        )


class Capitalize(PostProcessing):
    """ Applies capitalization strategies to content
    """
    NodeName = "capitalize"
    Marker: str = "🨁"  # NEUTRAL CHESS QUEEN
    RE_Upper: re.Regex = re.compile("(\p{Lu})")

    def __init__(self, first_word: Numeric, first_letters: Numeric,
                 column_token: str,
                 column_lemma: Optional[str] = None,
                 apply_unicode_marker: bool = False,
                 sentence_matcher: Optional[SentenceMatcherProto] = None):

        self.first_word: Numeric = first_word
        self.first_letters: Numeric = first_letters

        self.column_token: str = column_token
        self.column_lemma: Optional[str] = column_lemma
        self.apply_unicode_marker: bool = apply_unicode_marker
        self.sentence_matcher: Optional[SentenceMatcherProto] = sentence_matcher
        self.first_word_state: bool = True  # Variable representing the current status
        # (True = next word is a first word)

        self._files_chunks: Dict[str, List[bool]] = {}
        self._files_tokens: Dict[str, List[bool]] = {}

    @staticmethod
    def parse_when(value: str, ratio: Optional[str]) -> Numeric:
        if value == "always":
            return 1
        elif value == "never":
            return 0
        elif value == "random":
            return 0.5
        elif ratio:
            try:
                if 1.0 > float(ratio) > .0:
                    return float(ratio)
            except:
                raise ValueError("Your ration value is probably wrong. They must be < 1.0 (Found: {})".format(ratio))
        raise ValueError("Invalid parameters for a ratio or an application")

    @classmethod
    def parse_node_including_when(cls, node: Element, name: str) -> Tuple[Numeric, Optional[Element]]:
        target = node.findall("./{name}".format(name=name))
        if target:
            return cls.parse_when(target[0].attrib["when"], target[0].attrib.get("ratio")), target[0]
        return 0, None

    def _modify_line(self, header: List[str], values: Optional[List[str]],
                     file_path: str, config: "CorpusConfiguration") -> Dict[str, str]:
        if self.first_word and self.sentence_matcher.match(header, values):
            self.first_word_state = True
            if values:
                return dict(zip(header, values))

        if not values or len(header) != len(values):
            return {}

        line = dict(zip(header, values))

        # Sentence starts
        if self.first_word > .0 and self.first_word_state and self._files_chunks[file_path].pop():
            line[self.column_token] = line[self.column_token].capitalize()
            # Need to pop tokens as well
            if self.first_letters:
                self._files_tokens[file_path].pop()
        elif self.first_letters > .0 and self._files_tokens[file_path].pop():
            line[self.column_token] = line[self.column_token].capitalize()

        if self.apply_unicode_marker:
            line[self.column_token] = self.RE_Upper.sub(self._replace_caps, line[self.column_token])
            if self.column_lemma:
                line[self.column_lemma] = self.RE_Upper.sub(self._replace_caps, line[self.column_lemma])

        self.first_word_state = False
        return line

    def _replace_caps(self, value):
        return value.group().lower()+self.Marker

    @staticmethod
    def _transform_to_bool_list(count: int, ratio: Numeric) -> List[bool]:
        if ratio == 1.0:
            return [True] * count
        elif ratio == .0:
            return [False] * count
        else:
            positives = min(round(count * ratio), count)
            negatives = count - positives
            out = [True] * positives + [False] * negatives
            random.shuffle(out)
            return out

    def apply(self, file_path: str, config: "CorpusConfiguration"):
        # We scan the files
        chunks, tokens = self._scan_chunks(file_path, config, sentence_matcher=self.sentence_matcher)

        # We store the dispatch of booleans
        if self.first_word > .0:
            self._files_chunks[file_path] = self._transform_to_bool_list(chunks, self.first_word)
        if self.first_letters > .0:
            self._files_tokens[file_path] = self._transform_to_bool_list(tokens, self.first_letters)

        self._single_line_modify_routine(file_path=file_path, config=config)

    @classmethod
    def from_xml(cls, node: Element) -> "Capitalize":
        first_word, first_word_elem = cls.parse_node_including_when(node, "first-word")
        first_letters, _ = cls.parse_node_including_when(node, "first-letters")
        sentence_marker = None
        if first_word != .0:
            try:
                sentence_marker = SentenceMatcherProto.from_xml(
                    first_word_elem.findall("./sentence-marker")[0]
                )
            except IndexError:
                print("You forgot to use a sentence marker.")
                raise Exception

        return cls(
            first_word=first_word,
            first_letters=first_letters,
            sentence_matcher=sentence_marker,
            apply_unicode_marker=node.attrib.get("utf8-marker-for-caps", "false").lower() == "true",
            column_token=node.attrib["column-token"],
            column_lemma=node.attrib.get("column-lemma")
        )



import os
from glob import glob


def check_files(path):
    """ Check that there is files

    :param path: Path for a single or multiple files
    :return: List of files to process
    """
    if isinstance(path, (list, tuple)):
        return path
    elif os.path.isfile(path):
        data = [path]
    else:
        data = list(glob(path))
    if len(data):
        return data
    raise ValueError("No files were found at {}".format(path))


def check_ratio(train, test, dev):
    """ Check and adapt ratios from user input in case the sum is over 1.
    Current behavior is to scale done train.

    :param train: Training dataset ratio
    :param test: Testing dataset ratio
    :param dev: Dev dataset ratio
    :return: Train, test, dev
    """
    if train + test + dev > 1.0:
        train = 1.0 - test - dev
        print("Ratios are over 1, scaling down train ratio to {0:.2f}".format(train))
    return train, test, dev



from typing import List
import os


def get_name(output_folder, dataset, filename):
    return os.path.join(output_folder, dataset, os.path.basename(filename))


def add_sentence(
        output_folder: str, dataset: str, filename: str,
        sentence: List[str], source_marker: str, output_marker: str,
        subfolder: bool = True):
    """ Write a sentence in the given dataset

    :param output_folder:
    :param dataset:
    :param filename:
    :param sentence:
    :return:
    """
    if subfolder:
        filename = get_name(output_folder, dataset, filename)
    else:
        filename = get_name(output_folder, "", filename)
    if not os.path.isfile(filename):
        mode = "w"
    else:
        mode = "a"
    with open(filename, mode) as f:
        f.write("".join([s.replace(source_marker, output_marker) for s in sentence])+"\n")  # Add a secondary line break to keep things separated



__all__ = ["SentenceMatcherProto", "SentenceRegexpMatcher", "SentenceEmptyLineMatcher"]

from abc import ABC
from xml.etree.ElementTree import Element
from typing import Optional, List

import regex as re


class SentenceMatcherProto:
    def match(self, headers: List[str], values: Optional[List[str]]):
        raise NotImplementedError()

    @classmethod
    def from_xml(cls, node: Element) -> "SentenceMatcherProto":
        if node.attrib["name"] == "regexp":
            return SentenceRegexpMatcher.from_xml(node.xpath("./sentence-marker")[0])
        return SentenceEmptyLineMatcher()


class SentenceRegexpMatcher(SentenceMatcherProto):
    def __init__(self, regexp: re.Regex, column: str):
        self.regexp: re.Regex = regexp
        self.column: str = column

    def match(self, headers: List[str], values: Optional[List[str]]):
        data = dict(zip(headers, values))
        return self.regexp.match(data[self.column]) is not None

    @classmethod
    def from_xml(cls, node: Element) -> "SentenceRegexpMatcher":
        return cls(
            regexp=re.compile(node.attrib["regexp"]),
            column=node.attrib["column"]
        )


class SentenceEmptyLineMatcher(SentenceMatcherProto):
    def __init__(self):
        pass

    def match(self, headers: List[str], values: Optional[List[str]]):
        return not values or len(headers) != len(values)
